%%%----------------------------------------------------------------------
%%% File    : iconv.erl
%%% Author  : Alexey Shchepin <alexey@sevcom.net>
%%% Purpose : Interface to libiconv
%%% Created : 16 Feb 2003 by Alexey Shchepin <alexey@sevcom.net>
%%% Id      : $Id: iconv.erl 427 2005-10-25 02:19:01Z alexey $
%%%----------------------------------------------------------------------

-module(gen_linkedin_driver).
-author('echouzhou@tencent.com').
-vsn('$Revision: 427 $ ').

-behaviour(gen_server).

% API
-export([behaviour_info/1, start/1, restart/1, start_link/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, code_change/3, terminate/2]).
-export([control/3]).

-record(state, {module, driver_info}).

-define(PORT_COUNT, 1).

behaviour_info(callbacks) ->
	[{driver_info,0}];
behaviour_info(_Other) ->
	undefined.

start(Mod) ->
    gen_server:start({local, Mod}, ?MODULE, [Mod], []).

restart(Mod) ->
    gen_server:call(Mod, {restart}).

start_link(Mod) ->
    gen_server:start_link({local, Mod}, ?MODULE, [Mod], []).

get_lib_path() ->
	%app_util:get_lib_path().
	"priv/lib".

init([CallbackModule]) ->
	DriverInfo = CallbackModule:driver_info(),
	{DriverName, DriverTag, DriverTable} = DriverInfo,
	case erl_ddll:load_driver(get_lib_path(), DriverName) of 
		ok ->
	    	ets:new(DriverTable, [set, public, named_table]),
	    	lists:foreach(fun(I) ->
			                    Port = open_port({spawn_driver, DriverName}, []),
			                    ets:insert(DriverTable, {{DriverTag, I}, Port})
			                end, 
			                lists:seq(1, ?PORT_COUNT)),
	    	{ok, #state{module=CallbackModule, driver_info=DriverInfo}};	
		{error, ErrorDesc} ->
			io:format("[gen_linkedin_driver] init([~p]) ERROR because: ~p~n~p~n", [CallbackModule, ErrorDesc, erl_ddll:format_error(ErrorDesc)]),
			{stop, ErrorDesc}
	end.


%%% --------------------------------------------------------
%%% The call-back functions.
%%% --------------------------------------------------------

handle_call({restart}, _From, State) ->
	{DriverName, DriverTag, DriverTable} = State#state.driver_info,
	case ets:match(DriverTable, {{DriverTag, '_' }, '$1'}) of
        [] ->
            ok;
        PortNames ->
            io:format("[gen_linkedin_driver] Restarting all ~p : ~p \n", [DriverName, PortNames]),
            [port_close(P) || [P] <- PortNames],									%close all ports

			case erl_ddll:reload_driver(get_lib_path(), DriverName) of 		%reload the driver
				ok ->
			    	lists:foreach(fun(I) ->
					                    Port = open_port({spawn_driver, DriverName}, []),
					                    ets:insert(DriverTable, {{DriverTag, I}, Port})
					                end, 
					                lists:seq(1, ?PORT_COUNT));
				{error, ErrorDesc} ->
					io:format("[gen_linkedin_driver] restart([~p]) ERROR because: ~p~n~p~n", [DriverName, ErrorDesc, erl_ddll:format_error(ErrorDesc)])
			end,
            ok
    end,	
    {reply, ok, State};

handle_call(_, _, State) ->
    {noreply, State}.

handle_cast(_, State) ->
    {noreply, State}.

handle_info({'EXIT', Port, Reason}, State) ->
	{DriverName, _, DriverTable} = State#state.driver_info,
    case ets:match(DriverTable, {'$1', Port}) of
        [] ->
            io:format("[gen_linkedin_driver] Port ~p died: ~p\n", [Port, Reason]),
            ok;
        [[PortName]|_] ->
            io:format("[gen_linkedin_driver] Port died because ~p, will add a new one ...\n", [Reason]),
            NewPort = open_port({spawn_driver, DriverName}, []),
            ets:insert(DriverTable, {PortName, NewPort});
        _Any ->
            io:format("[gen_linkedin_driver] Huh, unexpected!? : ~p\n", [_Any]),
            ok
    end,
    {noreply, State};

handle_info(_, State) ->
    {noreply, State}.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

terminate(_Reason, State) ->
	{_, _, DriverTable} = State#state.driver_info,
    	Ports = ets:tab2list(DriverTable),
    	lists:foreach(fun(_PortName, Port) -> port_close(Port) end, Ports),
    	ok.


get_driver_port(CallbackModule) ->
	{_DriverName, DriverTag, DriverTable} = CallbackModule:driver_info(),
    	PortIndex = erlang:system_info(scheduler_id) rem ?PORT_COUNT + 1,
    	[{_, Port} | _] = ets:lookup(DriverTable, {DriverTag, PortIndex}),
    	Port.

% API 

control(Module, Cmd, Data) ->
    Port = get_driver_port(Module),
    erlang:port_control(Port, Cmd, Data).
